##
# types. - Package for I/O and PostgreSQL specific types.
##
"""
PostgreSQL types and identifiers.
"""
# XXX: Would be nicer to generate these from a header file...
InvalidOid = 0

RECORDOID = 2249
BOOLOID = 16
BITOID = 1560
VARBITOID = 1562
ACLITEMOID = 1033

CHAROID = 18
NAMEOID = 19
TEXTOID = 25
BYTEAOID = 17
BPCHAROID = 1042
VARCHAROID = 1043
CSTRINGOID = 2275
UNKNOWNOID = 705
REFCURSOROID = 1790
UUIDOID = 2950

TSVECTOROID = 3614
GTSVECTOROID = 3642
TSQUERYOID = 3615
REGCONFIGOID = 3734
REGDICTIONARYOID = 3769

JSONOID = 114
JSONBOID = 3802
XMLOID = 142

MACADDROID = 829
INETOID = 869
CIDROID = 650

TYPEOID = 71
PROCOID = 81
CLASSOID = 83
ATTRIBUTEOID = 75

DATEOID = 1082
TIMEOID = 1083
TIMESTAMPOID = 1114
TIMESTAMPTZOID = 1184
INTERVALOID = 1186
TIMETZOID = 1266
ABSTIMEOID = 702
RELTIMEOID = 703
TINTERVALOID = 704

INT8OID = 20
INT2OID = 21
INT4OID = 23
OIDOID = 26
TIDOID = 27
XIDOID = 28
CIDOID = 29
CASHOID = 790
FLOAT4OID = 700
FLOAT8OID = 701
NUMERICOID = 1700

POINTOID = 600
LINEOID = 628
LSEGOID = 601
PATHOID = 602
BOXOID = 603
POLYGONOID = 604
CIRCLEOID = 718

OIDVECTOROID = 30
INT2VECTOROID = 22
INT4ARRAYOID = 1007

REGPROCOID = 24
REGPROCEDUREOID = 2202
REGOPEROID = 2203
REGOPERATOROID = 2204
REGCLASSOID = 2205
REGTYPEOID = 2206
REGTYPEARRAYOID = 2211

TRIGGEROID = 2279
LANGUAGE_HANDLEROID = 2280
INTERNALOID = 2281
OPAQUEOID = 2282
VOIDOID = 2278
ANYARRAYOID = 2277
ANYELEMENTOID = 2283
ANYOID = 2276
ANYNONARRAYOID = 2776
ANYENUMOID = 3500

#: Mapping of type Oid to SQL type name.
oid_to_sql_name = {
	BPCHAROID : 'CHARACTER',
	VARCHAROID : 'CHARACTER VARYING',
	# *OID : 'CHARACTER LARGE OBJECT',

	# SELECT X'0F' -> bit. XXX: Does bytea have any play here?
	#BITOID : 'BINARY',
	#BYTEAOID : 'BINARY VARYING',
	# *OID : 'BINARY LARGE OBJECT',

	BOOLOID : 'BOOLEAN',

# exact numeric types
	INT2OID : 'SMALLINT',
	INT4OID : 'INTEGER',
	INT8OID : 'BIGINT',
	NUMERICOID : 'NUMERIC',

# approximate numeric types
	FLOAT4OID : 'REAL',
	FLOAT8OID : 'DOUBLE PRECISION',

# datetime types
	TIMEOID : 'TIME WITHOUT TIME ZONE',
	TIMETZOID : 'TIME WITH TIME ZONE',
	TIMESTAMPOID : 'TIMESTAMP WITHOUT TIME ZONE',
	TIMESTAMPTZOID : 'TIMESTAMP WITH TIME ZONE',
	DATEOID : 'DATE',

# interval types
	INTERVALOID : 'INTERVAL',

	XMLOID : 'XML',
}

#: Mapping of type Oid to name.
oid_to_name = {
	RECORDOID : 'record',
	BOOLOID : 'bool',
	BITOID : 'bit',
	VARBITOID : 'varbit',
	ACLITEMOID : 'aclitem',

	CHAROID : 'char',
	NAMEOID : 'name',
	TEXTOID : 'text',
	BYTEAOID : 'bytea',
	BPCHAROID : 'bpchar',
	VARCHAROID : 'varchar',
	CSTRINGOID : 'cstring',
	UNKNOWNOID : 'unknown',
	REFCURSOROID : 'refcursor',
	UUIDOID : 'uuid',

	TSVECTOROID : 'tsvector',
	GTSVECTOROID : 'gtsvector',
	TSQUERYOID : 'tsquery',
	REGCONFIGOID : 'regconfig',
	REGDICTIONARYOID : 'regdictionary',

	XMLOID : 'xml',
	JSONOID : 'json',
	JSONBOID : 'jsonb',

	MACADDROID : 'macaddr',
	INETOID : 'inet',
	CIDROID : 'cidr',

	TYPEOID : 'type',
	PROCOID : 'proc',
	CLASSOID : 'class',
	ATTRIBUTEOID : 'attribute',

	DATEOID : 'date',
	TIMEOID : 'time',
	TIMESTAMPOID : 'timestamp',
	TIMESTAMPTZOID : 'timestamptz',
	INTERVALOID : 'interval',
	TIMETZOID : 'timetz',
	ABSTIMEOID : 'abstime',
	RELTIMEOID : 'reltime',
	TINTERVALOID : 'tinterval',

	INT8OID : 'int8',
	INT2OID : 'int2',
	INT4OID : 'int4',
	OIDOID : 'oid',
	TIDOID : 'tid',
	XIDOID : 'xid',
	CIDOID : 'cid',
	CASHOID : 'cash',
	FLOAT4OID : 'float4',
	FLOAT8OID : 'float8',
	NUMERICOID : 'numeric',

	POINTOID : 'point',
	LINEOID : 'line',
	LSEGOID : 'lseg',
	PATHOID : 'path',
	BOXOID : 'box',
	POLYGONOID : 'polygon',
	CIRCLEOID : 'circle',

	OIDVECTOROID : 'oidvector',
	INT2VECTOROID : 'int2vector',
	INT4ARRAYOID : 'int4array',

	REGPROCOID : 'regproc',
	REGPROCEDUREOID : 'regprocedure',
	REGOPEROID : 'regoper',
	REGOPERATOROID : 'regoperator',
	REGCLASSOID : 'regclass',
	REGTYPEOID : 'regtype',
	REGTYPEARRAYOID : 'regtypearray',

	TRIGGEROID : 'trigger',
	LANGUAGE_HANDLEROID : 'language_handler',
	INTERNALOID : 'internal',
	OPAQUEOID : 'opaque',
	VOIDOID : 'void',
	ANYARRAYOID : 'anyarray',
	ANYELEMENTOID : 'anyelement',
	ANYOID : 'any',
	ANYNONARRAYOID : 'anynonarray',
	ANYENUMOID : 'anyenum',
}

name_to_oid = dict(
	[(v,k) for k,v in oid_to_name.items()]
)

class Array(object):
	"""
	Type used to mimic PostgreSQL arrays. While there are many semantic
	differences, the primary one is that the elements contained by an Array
	instance are not strongly typed. The purpose of this class is to provide
	some consistency with PostgreSQL with respect to the structure of an Array.

	The structure consists of three parts:

	 * The elements of the array.
	 * The lower boundaries.
	 * The upper boundaries.

	There is also a `dimensions` property, but it is derived from the
	`lowerbounds` and `upperbounds` to yield a normalized description of the
	ARRAY's structure.

	The Python interfaces, such as __getitem__, are *not* subjected to the
	semantics of the lower and upper bounds. Rather, the normalized dimensions
	provide the primary influence for these interfaces. So, unlike SQL
	indirection, getting an index that does *not* exist will raise a Python
	`IndexError`.
	"""
	# return an iterator over the absolute elements of a nested sequence
	@classmethod
	def unroll_nest(typ, hier, dimensions, depth = 0):
		dsize = dimensions and dimensions[depth] or 0
		if len(hier) != dsize:
			raise ValueError("list size not consistent with dimensions at depth " + str(depth))
		r = []
		ndepth = depth + 1
		if ndepth == len(dimensions):
			# at the bottom
			r = hier
		else:
			# go deeper
			for x in hier:
				r.extend(typ.unroll_nest(x, dimensions, ndepth))
		return r

	# Detect the dimensions of a nested sequence
	@staticmethod
	def detect_dimensions(hier, len = len):
		# if the list is empty, it's a zero-dimension array.
		if hier:
			yield len(hier)
			hier = hier[0]
			depth = 1
			while hier.__class__ is list:
				depth += 1
				l = len(hier)
				if l < 1:
					raise ValueError("axis {0} is empty".format(depth))
				yield l
				hier = hier[0]

	@classmethod
	def from_elements(typ,
		elements,
		lowerbounds = None,
		upperbounds = None,
		len = len,
	):
		"""
		Instantiate an Array from the given elements, lowerbounds, and upperbounds.

		The given elements are bound to the array which provides them with the
		structure defined by the lower boundaries and the upper boundaries.

		A `ValueError` will be raised in the following situations:

		 * The number of elements given are inconsistent with the number of elements
		   described by the upper and lower bounds.
		 * The lower bounds at a given axis exceeds the upper bounds at a given
		   axis.
		 * The number of lower bounds is inconsistent with the number of upper
		   bounds.
		"""
		# resolve iterable
		elements = list(elements)
		nelements = len(elements)

		# If ndims is zero, lowerbounds will be ()
		if lowerbounds is None:
			if upperbounds:
				lowerbounds = (1,) * len(upperbounds)
			elif nelements == 0:
				# special for empty ARRAY; no dimensions.
				lowerbounds = ()
			else:
				# one dimension.
				lowerbounds = (1,)
		else:
			lowerbounds = tuple(lowerbounds)

		if upperbounds is not None:
			upperbounds = tuple(upperbounds)
			dimensions = []
			# upperbounds were given, so check.
			if upperbounds:
				elcount = 1
				for lb, ub in zip(lowerbounds, upperbounds):
					x = ub - lb + 1
					if x < 1:
						# special case empty ARRAYs
						if nelements == 0:
							upperbounds = ()
							lowerbounds = ()
							dimensions = ()
							elcount = 0
							break
						raise ValueError("lowerbounds exceeds upperbounds")
					# physical dimensions.
					dimensions.append(x)
					elcount = x * elcount
			else:
				elcount = 0
			if nelements != elcount:
				raise ValueError("element count inconsistent with boundaries")
			dimensions = tuple(dimensions)
		else:
			# fill in default
			if nelements == 0:
				upperbounds = ()
				dimensions = ()
			else:
				upperbounds = (nelements,)
				dimensions = (nelements,)

		# consistency..
		if len(lowerbounds) != len(upperbounds):
			raise ValueError("number of lowerbounds inconsistent with upperbounds")

		rob = super().__new__(typ)
		rob._elements = elements
		rob.lowerbounds = lowerbounds
		rob.upperbounds = upperbounds
		rob.dimensions = dimensions
		rob.ndims = len(dimensions)
		rob._weight = len(rob._elements) // (dimensions and dimensions[0] or 1)
		return rob

	# Method used to create an Array() from nested lists.
	@classmethod
	def from_nest(typ, nest):
		dims = tuple(typ.detect_dimensions(nest))
		return typ.from_elements(
			list(typ.unroll_nest(nest, dims)),
			upperbounds = dims,
			# lowerbounds is implied to (1,)*len(upper)
		)

	def __new__(typ, nested_elements):
		"""
		Create an types.Array() using the given nested lists. The boundaries of
		the array are detected by traversing the first items of the nested
		lists::

			Array([[1,2,4],[3,4,8]])

		Lists are used to define the boundaries so that tuples may be used to
		represent any complex elements. The above array will the `lowerbounds`
		``(1,1)``, and the `upperbounds` ``(2,3)``.
		"""
		if nested_elements.__class__ is Array:
			return nested_elements
		return typ.from_nest(list(nested_elements))

	def __getnewargs__(self):
		return (self.nest(),)

	def elements(self):
		"""
		Returns an iterator to the elements of the Array. The elements are
		produced in physical order.
		"""
		return iter(self._elements)

	def nest(self, seqtype = list):
		"""
		Transform the array into a nested list.

		The `seqtype` keyword can be used to override the type used to represent
		the elements of a given axis.
		"""
		if self.ndims < 2:
			return seqtype(self._elements)
		else:
			rl = []
			for x in self:
				rl.append(x.nest(seqtype = seqtype))
			return seqtype(rl)

	def get_element(self, address,
		idxerr = "index {0} at axis {1} is out of range {2}".format
	):
		"""
		Get an element in the array using the given axis sequence.

		>>> a=Array([[1,2],[3,4]])
		>>> a.get_element((0,0)) == 1
		True
		>>> a.get_element((1,1)) == 4
		True

		This is similar to getting items in a nested list::

		>>> l=[[1,2],[3,4]]
		>>> l[0][0] == 1
		True
		"""
		if not self.dimensions:
			raise IndexError("array is empty")
		if len(address) != len(self.dimensions):
			raise ValueError("given axis sequence is inconsistent with number of dimensions")

		# normalize axis specification (-N + DIM), check for IndexErrors, and
		# resolve the element's position.
		cur = 0
		nelements = len(self._elements)
		for n, a, dim in zip(range(len(address)), address, self.dimensions):
			if a < 0:
				a = a + dim
				if a < 0:
					raise IndexError(idxerr(a, n, dim))
			else:
				if a >= dim:
					raise IndexError(idxerr(a, n, dim))
			nelements = nelements // dim
			cur += (a * nelements)
		return self._elements[cur]

	def sql_get_element(self, address):
		"""
		Like `get_element`, but with SQL indirection semantics. Notably, returns
		`None` on IndexError.
		"""
		try:
			a = [a - lb for (a, lb) in zip(address, self.lowerbounds)]
			# get_element accepts negatives, so check the converted sequence.
			for x in a:
				if x < 0:
					return None
			return self.get_element(a)
		except IndexError:
			return None

	def __repr__(self):
		return '%s.%s(%r)' %(
			type(self).__module__,
			type(self).__name__,
			self.nest()
		)

	def __len__(self):
		return self.dimensions and self.dimensions[0] or 0

	def __eq__(self, ob):
		return list(self) == ob

	def __ne__(self, ob):
		return list(self) != ob

	def __gt__(self, ob):
		return list(self) > ob

	def __lt__(self, ob):
		return list(self) < ob

	def __le__(self, ob):
		return list(self) <= ob

	def __ge__(self, ob):
		return list(self) >= ob

	def __getitem__(self, item):
		if self.ndims < 2:
			# Array with 1dim is more or less a list.
			return self._elements[item]
		if isinstance(item, slice):
			# get a sub-array slice
			l = len(self)
			n = 0
			r = []
			# for each offset in the slice, get the elements and add them
			# to the new elements list used to build the new Array().
			for x in range(*(item.indices(l))):
				n = n + 1
				r.extend(
					self._elements[slice(self._weight*x,self._weight*(x+1))]
				)
			if n:
				return self.__class__.from_elements(r,
					lowerbounds = (1,) + self.lowerbounds[1:],
					upperbounds = (n,) + self.upperbounds[1:],
				)
			else:
				# Empty
				return self.__class__.from_elements(())
		else:
			# get a sub-array
			l = len(self)
			if item > l:
				raise IndexError("index {0} is out of range".format(l))
			return self.__class__.from_elements(
				self._elements[self._weight*item:self._weight*(item+1)],
				lowerbounds = self.lowerbounds[1:],
				upperbounds = self.upperbounds[1:],
			)

	def __iter__(self):
		if self.ndims < 2:
			# Special case empty and single dimensional ARRAYs
			return self.elements()
		return (self[x] for x in range(len(self)))

from operator import itemgetter
get0 = itemgetter(0)
get1 = itemgetter(1)
del itemgetter

class Row(tuple):
	"Name addressable items tuple; mapping and sequence"
	@classmethod
	def from_mapping(typ, keymap, map, get1 = get1):
		iter = [
			map.get(k) for k,_ in sorted(keymap.items(), key = get1)
		]
		r = typ(iter)
		r.keymap = keymap
		return r

	@classmethod
	def from_sequence(typ, keymap, seq):
		r = typ(seq)
		r.keymap = keymap
		return r

	def __getitem__(self, i, gi = tuple.__getitem__):
		if isinstance(i, (int, slice)):
			return gi(self, i)
		idx = self.keymap[i]
		return gi(self, idx)

	def get(self, i, gi = tuple.__getitem__, len = len):
		if type(i) is int:
			l = len(self)
			if -l < i < l:
				return gi(self, i)
		else:
			idx = self.keymap.get(i)
			if idx is not None:
				return gi(self, idx)
		return None

	def keys(self):
		return self.keymap.keys()

	def values(self):
		return iter(self)

	def items(self):
		return zip(iter(self.column_names), iter(self))

	def index_from_key(self, key):
		return self.keymap.get(key)

	def key_from_index(self, index):
		for k,v in self.keymap.items():
			if v == index:
				return k
		return None

	@property
	def column_names(self, get0 = get0, get1 = get1):
		l=list(self.keymap.items())
		l.sort(key=get1)
		return tuple(map(get0, l))

	def transform(self, *args, **kw):
		"""
		Make a new Row after processing the values with the callables associated
		with the values either by index, \*args, or my column name, \*\*kw.

			>>> r=Row.from_sequence({'col1':0,'col2':1}, (1,'two'))
			>>> r.transform(str)
			('1','two')
			>>> r.transform(col2 = str.upper)
			(1,'TWO')
			>>> r.transform(str, col2 = str.upper)
			('1','TWO')

		Combine with methodcaller and map to transform lots of rows:

			>>> rowseq = [r]
			>>> xf = operator.methodcaller('transform', col2 = str.upper)
			>>> list(map(xf, rowseq))
			[(1,'TWO')]

		"""
		r = list(self)
		i = 0
		for x in args:
			if x is not None:
				r[i] = x(tuple.__getitem__(self, i))
			i = i + 1
		for k,v in kw.items():
			if v is not None:
				i = self.index_from_key(k)
				if i is None:
					raise KeyError("row has no such key, " + repr(k))
				r[i] = v(self[k])
		return type(self).from_sequence(self.keymap, r)



##
# .types.namedtuple - return rows as namedtuples
##
"""
Factories for namedtuple row representation.
"""
from collections import namedtuple

#: Global namedtuple type cache.
cache = {}

# Build and cache the namedtuple's produced.
def _factory(colnames : [str], namedtuple = namedtuple) -> tuple:
	global cache
	# Provide some normalization.
	# Anything beyond this can just get renamed.
	colnames = tuple([
		x.replace(' ', '_') for x in colnames
	])
	try:
		return cache[colnames]
	except KeyError:
		NT = namedtuple('row', colnames, rename = True)
		cache[colnames] = NT
		return NT

def NamedTupleFactory(attribute_map, composite_relid = None):
	"""
	Alternative db.typio.RowFactory for producing namedtuple's instead of
	postgresql.types.Row() instances.

	To install::

		>>> from py_opengauss.types.namedtuple import NamedTupleFactory
		>>> import py_opengauss
		>>> db = py_opengauss.open(...)
		>>> db.typio.RowTypeFactory(NamedTupleFactory)
	
	And **all** Rows produced by that connection will be namedtuple()'s.
	This includes composites.
	"""
	colnames = list(attribute_map.items())
	colnames.sort(key = lambda x: x[1])
	return lambda y: _factory((x[0] for x in colnames))(*y)

from itertools import chain, starmap

def namedtuples(stmt, from_iter = chain.from_iterable, map = starmap):
	"""
	Alternative to the .rows() execution method.

	Use::
	
		>>> from py_opengauss.types.namedtuple import namedtuples
		>>> ps = namedtuples(db.prepare(...))
		>>> for nt in ps(...):
		...  nt.a_column_name

	This effectively selects the execution method to be used with the statement.
	"""
	NT = _factory(stmt.column_names)
	# build the execution "method"
	chunks = stmt.chunks
	def rows_as_namedtuples(*args, **kw):
		return map(NT, from_iter(chunks(*args, **kw))) # starmap
	return rows_as_namedtuples

del chain, starmap



##
# .test.test_lib - test the .lib package
##
import sys
import os
import unittest
import tempfile

from .. import exceptions as pg_exc
from .. import lib as pg_lib
from .. import sys as pg_sys
from ..temporal import pg_tmp

ilf = """
preface

[sym]
select 1
[sym_ref]
*[sym]
[sym_ref_trail]
*[sym] WHERE FALSE
[sym_first::first]
select 1


[sym_rows::rows]
select 1

[sym_chunks::chunks]
select 1

[sym_declare::declare]
select 1

[sym_const:const:first]
select 1
[sym_const_rows:const:rows]
select 1
[sym_const_chunks:const:chunks]
select 1
[sym_const_column:const:column]
select 1
[sym_const_ddl:const:]
create temp table sym_const_dll (i int);

[sym_preload:preload:first]
select 1

[sym_proc:proc]
test_ilf_proc(int)

[sym_srf_proc:proc]
test_ilf_srf_proc(int)

[&sym_reference]
SELECT 'SELECT 1';

[&sym_reference_params]
SELECT 'SELECT ' || $1::text;

[&sym_reference_first::first]
SELECT 'SELECT 1::int4';

[&sym_reference_const:const:first]
SELECT 'SELECT 1::int4';

[&sym_reference_proc:proc]
SELECT 'test_ilf_proc(int)'::text
"""

class test_lib(unittest.TestCase):
	# NOTE: Module libraries are implicitly tested
	# in postgresql.test.test_driver; much functionality
	# depends on the `sys` library.
	def _testILF(self, lib):
		self.assertTrue('preface' in lib.preface)
		db.execute("CREATE OR REPLACE FUNCTION test_ilf_proc(int) RETURNS int language sql as 'select $1';")
		db.execute("CREATE OR REPLACE FUNCTION test_ilf_srf_proc(int) RETURNS SETOF int language sql as 'select $1';")
		b = pg_lib.Binding(db, lib)
		self.assertEqual(b.sym_ref(), [(1,)])
		self.assertEqual(b.sym_ref_trail(), [])
		self.assertEqual(b.sym(), [(1,)])
		self.assertEqual(b.sym_first(), 1)
		self.assertEqual(list(b.sym_rows()), [(1,)])
		self.assertEqual([list(x) for x in b.sym_chunks()], [[(1,)]])
		c = b.sym_declare()
		self.assertEqual(c.read(), [(1,)])
		c.seek(0)
		self.assertEqual(c.read(), [(1,)])
		self.assertEqual(b.sym_const, 1)
		self.assertEqual(b.sym_const_column, [1])
		self.assertEqual(b.sym_const_rows, [(1,)])
		self.assertEqual(b.sym_const_chunks, [[(1,)]])
		self.assertEqual(b.sym_const_ddl, ('CREATE TABLE', None))
		self.assertEqual(b.sym_preload(), 1)
		# now stored procs
		self.assertEqual(b.sym_proc(2,), 2)
		self.assertEqual(list(b.sym_srf_proc(2,)), [2])
		self.assertRaises(AttributeError, getattr, b, 'LIES')
		# reference symbols
		self.assertEqual(b.sym_reference()(), [(1,)])
		self.assertEqual(b.sym_reference_params('1::int')(), [(1,)])
		self.assertEqual(b.sym_reference_params("'foo'::text")(), [('foo',)])
		self.assertEqual(b.sym_reference_first()(), 1)
		self.assertEqual(b.sym_reference_const(), 1)
		self.assertEqual(b.sym_reference_proc()(2,), 2)

	@pg_tmp
	def testILF_from_lines(self):
		lib = pg_lib.ILF.from_lines([l + '\n' for l in ilf.splitlines()])
		self._testILF(lib)

	@pg_tmp
	def testILF_from_file(self):
		f = tempfile.NamedTemporaryFile(
			delete = False, mode = 'w', encoding = 'utf-8'
		) 
		n = f.name
		try:
			f.write(ilf)
			f.flush()
			f.seek(0)
			lib = pg_lib.ILF.open(n, encoding = 'utf-8')
			self._testILF(lib)
			f.close()
		finally:
			# so annoying...
			os.unlink(n)

	@pg_tmp
	def testLoad(self):
		# gotta test it in the cwd...
		pid = os.getpid()
		frag = 'temp' + str(pid)
		fn = 'lib' + frag + '.sql'
		try:
			with open(fn, 'w') as f:
				f.write("[foo]\nSELECT 1")
			pg_sys.libpath.insert(0, os.path.curdir)
			l = pg_lib.load(frag)
			b = pg_lib.Binding(db, l)
			self.assertEqual(b.foo(), [(1,)])
		finally:
			os.remove(fn)
	
	@pg_tmp
	def testCategory(self):
		lib = pg_lib.ILF.from_lines([l + '\n' for l in ilf.splitlines()])
		# XXX: evil, careful..
		lib._name = 'name'
		c = pg_lib.Category(lib)
		c(db)
		self.assertEqual(db.name.sym_first(), 1)
		c = pg_lib.Category(renamed = lib)
		c(db)
		self.assertEqual(db.renamed.sym_first(), 1)

	@pg_tmp
	def testCategoryAliases(self):
		lib = pg_lib.ILF.from_lines([l + '\n' for l in ilf.splitlines()])
		# XXX: evil, careful..
		lib._name = 'name'
		c = pg_lib.Category(lib, renamed = lib)
		c(db)
		self.assertEqual(db.name.sym_first(), 1)
		self.assertEqual(db.renamed.sym_first(), 1)

if __name__ == '__main__':
	unittest.main()



from .. import INETOID, CIDROID, MACADDROID
from . import lib
import ipaddress

oid_to_type = {
	MACADDROID : str,
	INETOID: ipaddress._IPAddressBase,
	CIDROID: ipaddress._BaseNetwork,
}

def inet_pack(ob, pack = lib.net_pack, Constructor = ipaddress.ip_address):
	a = Constructor(ob)
	return pack((a.version, None, a.packed))

def cidr_pack(ob, pack = lib.net_pack, Constructor = ipaddress.ip_network):
	a = Constructor(ob)
	return pack((a.version, a.prefixlen, a.network_address.packed))

def inet_unpack(data, unpack = lib.net_unpack, Constructor = ipaddress.ip_address):
	version, mask, data = unpack(data)
	return Constructor(data)

def cidr_unpack(data, unpack = lib.net_unpack, Constructor = ipaddress.ip_network):
	version, mask, data = unpack(data)
	return Constructor(data).supernet(new_prefix=mask)

oid_to_io = {
	MACADDROID : (lib.macaddr_pack, lib.macaddr_unpack, str),
	CIDROID : (cidr_pack, cidr_unpack, str),
	INETOID : (inet_pack, inet_unpack, str),
}



##
# py-postgresql root package
# http://github.com/python-postgres/fe
##
"""
py-postgresql is a Python package for using PostgreSQL. This includes low-level
protocol tools, a driver(PG-API and DB-API 2.0), and cluster management tools.

See <http://postgresql.org> for more information about PostgreSQL and <http://python.org>
for information about Python.
"""
__all__ = [
	'__author__',
	'__date__',
	'__version__',
	'__docformat__',
	'version',
	'version_info',
	'open',
]

#: The version string of py-postgresql.
version = '' # overridden by subsequent import from .project.

#: The version triple of py-postgresql: (major, minor, patch).
version_info = () # overridden by subsequent import from .project.

# Optional.
try:
	from .project import version_info, version, \
		author as __author__, date as __date__
	__version__ = version
except ImportError:
	pass

# Check that the given connection is the primary instance
def is_primary(c):
	sql = "SELECT local_role,db_state FROM pg_stat_get_stream_replications()"
	r = c.prepare(sql)()
	if r:
		# 主备实例时角色为 Primary，单实例时为 Normal
		if r[0][0] in ('Primary', 'Normal') and r[0][1] == 'Normal':
			return True
	return False

# Avoid importing these until requested.
_pg_iri = _pg_driver = _pg_param = None
def open(iri = None, prompt_title = None, **kw):
	"""
	Create a `postgresql.api.Connection` to the server referenced by the given
	`iri`::

		>>> import py_opengauss
		# General Format:
		>>> db = py_opengauss.open('pq://user:password@host:port/database')

		# Also support opengauss scheme:
		>>> db = py_opengauss.open('opengauss://user:password@host:port/database')

		# multi IP support:
		>>> db = py_opengauss.open('opengauss://user:password@host1:123,host2:456/database')

		# Connect to 'postgres' at localhost.
		>>> db = py_opengauss.open('localhost/postgres')

	Connection keywords can also be used with `open`. See the narratives for
	more information.

	The `prompt_title` keyword is ignored. `open` will never prompt for
	the password unless it is explicitly instructed to do so.

	(Note: "pq" is the name of the protocol used to communicate with PostgreSQL)
	"""
	global _pg_iri, _pg_driver, _pg_param
	if _pg_iri is None:
		from . import iri as _pg_iri
		from . import driver as _pg_driver
		from . import clientparameters as _pg_param

	return_connector = False
	if iri is not None:
		if iri.startswith('&'):
			return_connector = True
			iri = iri[1:]
		iri_params = _pg_iri.parse(iri)
		[p.pop('path', None) for p in iri_params]
	else:
		iri_params = []

	std_params = _pg_param.collect(prompt_title = None)

	# Traversal connect host for search primary
	errs = []
	for iri_param in iri_params:
		# If unix is specified, it's going to conflict with any standard
		# settings, so remove them right here.
		if 'unix' in kw or 'unix' in iri_param:
			std_params.pop('host', None)
			std_params.pop('port', None)
		params = _pg_param.normalize(
			list(_pg_param.denormalize_parameters(std_params)) + \
			list(_pg_param.denormalize_parameters(iri_param)) + \
			list(_pg_param.denormalize_parameters(kw))
		)
		_pg_param.resolve_password(params)

		C = _pg_driver.default.fit(**params)
		c = C()
		if len(iri_params) == 1:
			if return_connector:
				return C
			else:
				c.connect()
				return c
		try:
			c.connect()
		except Exception as e:
			errs.append({params.get('host'): e})
			continue
		if is_primary(c):
			return C if return_connector is True else c
		else:
			c.close()
			errs.append({params.get('host'): "not primary instance"})

	raise ConnectionError(errs)

__docformat__ = 'reStructuredText'



from ...types import OIDOID, XIDOID, CIDOID, TIDOID
from . import lib

oid_to_io = {
	OIDOID : (lib.oid_pack, lib.oid_unpack),
	XIDOID : (lib.xid_pack, lib.xid_unpack),
	CIDOID : (lib.cid_pack, lib.cid_unpack),
	TIDOID : (lib.tid_pack, lib.tid_unpack),
	#ACLITEMOID : (aclitem_pack, aclitem_unpack),
}



##
# .temporal - manage the temporary cluster
##
"""
Temporary PostgreSQL cluster for the process.
"""
import os
import atexit
import builtins
from collections import deque
from .cluster import Cluster, ClusterError
from . import installation
from .python.socket import find_available_port

class Temporal(object):
	"""
	Manages a temporary cluster for the duration of the process.

	Instances of this class reference a distinct cluster. These clusters are
	transient; they will only exist until the process exits.

	Usage::

		>>> from py_opengauss.temporal import pg_tmp
		>>> with pg_tmp:
		...  ps = db.prepare('SELECT 1')
		...  assert ps.first() == 1

	Or `pg_tmp` can decorate a method or function.
	"""

	format_sandbox_id = staticmethod(('sandbox{0}_{1}').format)
	cluster_dirname = staticmethod(('pg_tmp_{0}_{1}').format)
	cluster = None

	_init_pid_ = None
	_local_id_ = 0
	builtins_keys = {
		'connector',
		'db',
		'do',
		'xact',
		'proc',
		'settings',
		'prepare',
		'sqlexec',
		'newdb',
	}

	def __init__(self):
		self.builtins_stack = deque()
		self.sandbox_id = 0
		# identifier for keeping temporary instances unique.
		self.__class__._local_id_ = self.local_id = (self.__class__._local_id_ + 1)

	def __call__(self, callable):
		def in_pg_temporal_context(*args, **kw):
			with self:
				return callable(*args, **kw)
		n = getattr(callable, '__name__', None)
		if n:
			in_pg_temporal_context.__name__ = n
		return in_pg_temporal_context

	def destroy(self):
		# Don't destroy if it's not the initializing process.
		if os.getpid() == self._init_pid_:
			# Kill all the open connections.
			try:
				c = cluster.connection(user = 'test', database = 'template1',)
				with c:
					if c.version_info[:2] <= (9,1):
						c.sys.terminate_backends()
					else:
						c.sys.terminate_backends_92()
			except Exception:
				# Doesn't matter much if it fails.
				pass
			cluster = self.cluster
			self.cluster = None
			self._init_pid_ = None
			if cluster is not None:
				cluster.stop()
				cluster.wait_until_stopped(timeout = 5)
				cluster.drop()

	def init(self,
		installation_factory = installation.default,
		inshint = {
			'hint' : "Try setting the PGINSTALLATION " \
			"environment variable to the `pg_config` path"
		}
	):
		if self.cluster is not None or 'PGTEST' in os.environ:
			return
		##
		# Hasn't been created yet, but doesn't matter.
		# On exit, obliterate the cluster directory.
		self._init_pid_ = os.getpid()
		atexit.register(self.destroy)

		# [$HOME|.]/.pg_tmpdb_{pid}
		self.cluster_path = os.path.join(
			os.environ.get('HOME', os.getcwd()),
			self.cluster_dirname(self._init_pid_, self.local_id)
		)
		self.logfile = os.path.join(self.cluster_path, 'logfile')
		installation = installation_factory()
		if installation is None:
			raise ClusterError(
				'could not find the default pg_config', details = inshint
			)

		vi = installation.version_info
		cluster = Cluster(installation, self.cluster_path,)

		# If it exists already, destroy it.
		if cluster.initialized():
			cluster.drop()
		cluster.encoding = 'utf-8'
		cluster.init(
			user = 'test', # Consistent username.
			encoding = cluster.encoding,
			logfile = None,
		)

		try:
			self.cluster_port = find_available_port()
		except:
			# Rely on chain.
			raise ClusterError(
				'could not find a port for the test cluster on localhost',
				creator = cluster
			)

		if vi[:2] > (9,6):
			# Default changed in 10.x
			cluster.settings['max_wal_senders'] = '0'

		cluster.settings.update(dict(
			port = str(self.cluster_port),
			max_connections = '20',
			shared_buffers = '200',
			listen_addresses = 'localhost',
			log_destination = 'stderr',
			log_min_messages = 'FATAL',
			max_prepared_transactions = '10',
		))

		if installation.version_info[:2] < (9, 3):
			cluster.settings.update(dict(
				unix_socket_directory = cluster.data_directory,
			))
		else:
			cluster.settings.update(dict(
				unix_socket_directories = cluster.data_directory,
			))

		# Start the database cluster.
		with open(self.logfile, 'w') as lfo:
			cluster.start(logfile = lfo)
		cluster.wait_until_started()

		# Initialize template1 and the test user database.
		c = cluster.connection(user = 'test', database = 'template1',)
		with c:
			c.execute('create database test')
		self.cluster = cluster

	def push(self):
		if 'PGTEST' in os.environ:
			from . import open as pg_open
			c = pg_open(os.environ['PGTEST']) # Ignoring PGINSTALLATION.
		else:
			c = self.cluster.connection(user = 'test')
			c.connect()

		extras = []
		sbid = self.format_sandbox_id(os.getpid(), self.sandbox_id + 1)

		def new_pg_tmp_connection(l = extras, clone = c.clone, sbid = sbid):
			# Used to create a new connection that will be closed
			# when the context stack is popped along with 'db'.
			l.append(clone())
			l[-1].settings['search_path'] = str(sbid) + ',' + l[-1].settings['search_path']
			return l[-1]

		# The new builtins.
		local_builtins = {
			'db' : c,
			'prepare' : c.prepare,
			'xact' : c.xact,
			'sqlexec' : c.execute,
			'do' : c.do,
			'settings' : c.settings,
			'proc' : c.proc,
			'connector' : c.connector,
			'new' : new_pg_tmp_connection,
		}
		if not self.builtins_stack:
			# Store any of those set or not set.
			current = {
				k : builtins.__dict__[k] for k in self.builtins_keys
				if k in builtins.__dict__
			}
			self.builtins_stack.append((current, []))

		# Store and push.
		self.builtins_stack.append((local_builtins, extras))
		builtins.__dict__.update(local_builtins)
		self.sandbox_id += 1

	def pop(self, exc, drop_schema = ('DROP SCHEMA {0} CASCADE').format):
		local_builtins, extras = self.builtins_stack.pop()
		self.sandbox_id -= 1

		# restore __builtins__
		if len(self.builtins_stack) > 1:
			builtins.__dict__.update(self.builtins_stack[-1][0])
		else:
			previous = self.builtins_stack.popleft()
			for x in self.builtins_keys:
				if x in previous:
					builtins.__dict__[x] = previous[x]
				else:
					# Wasn't set before.
					builtins.__dict__.pop(x, None)

		# close popped connection, but only if we're not in an interrupt.
		# However, temporal will always terminate all backends atexit.
		if exc is None or isinstance(exc, Exception):
			# Interrupt then close. Just in case something is lingering.
			for xdb in [local_builtins['db']] + list(extras):
				if xdb.closed is False:
					# In order for a clean close of the connection,
					# interrupt before closing. It is still
					# possible for the close to block, but less likely.
					xdb.interrupt()
				xdb.close()

			# Interrupted and closed all the other connections at this level;
			# now remove the sandbox schema.
			xdb = local_builtins['db']
			with xdb.clone() as c:
				# Use a new connection so that the state of
				# the context connection will not have to be
				# contended with.
				c.execute(drop_schema(self.format_sandbox_id(os.getpid(), self.sandbox_id + 1)))
		else:
			# interrupt exception; avoid waiting for close
			pass

	def _init_c(self, cxn):
		cxn.connect()
		sb = self.format_sandbox_id(os.getpid(), self.sandbox_id)
		cxn.execute('CREATE SCHEMA ' + sb)
		cxn.settings['search_path'] = ','.join((sb, cxn.settings['search_path']))

	def __enter__(self):
		if self.cluster is None:
			self.init()

		self.push()
		try:
			self._init_c(builtins.db)
		except Exception as e:
			# failed to initialize sandbox schema; pop it.
			self.pop(e)
			raise

	def __exit__(self, exc, val, tb):
		self.pop(val)

#: The process' temporary cluster or connection source.
pg_tmp = Temporal()



##
# .test.test_cluster
##
import sys
import os
import time
import unittest
import tempfile
from .. import installation
from ..cluster import Cluster, ClusterStartupError

default_installation = installation.default()

class test_cluster(unittest.TestCase):
	def setUp(self):
		self.cluster = Cluster(default_installation, 'test_cluster',)

	def tearDown(self):
		if self.cluster.installation is not None:
			self.cluster.drop()
			self.cluster = None

	def start_cluster(self, logfile = None):
		self.cluster.start(logfile = logfile)
		self.cluster.wait_until_started(timeout = 10)

	def init(self, *args, **kw):
		self.cluster.init(*args, **kw)

		vi = self.cluster.installation.version_info[:2]
		if vi >= (9, 3):
			usd = 'unix_socket_directories'
		else:
			usd = 'unix_socket_directory'

		if vi > (9, 6):
			self.cluster.settings['max_wal_senders'] = '0'

		self.cluster.settings.update({
			'max_connections' : '8',
			'listen_addresses' : 'localhost',
			'port' : '6543',
			usd : self.cluster.data_directory,
		})

	@unittest.skipIf(default_installation is None, "no installation provided by environment")
	def testSilentMode(self):
		self.init()
		self.cluster.settings['silent_mode'] = 'on'
		# if it fails to start(ClusterError), silent_mode is not working properly.
		try:
			self.start_cluster(logfile = sys.stdout)
		except ClusterStartupError:
			# silent_mode is not supported on windows by PG.
			if sys.platform in ('win32','win64'):
				pass
			elif self.cluster.installation.version_info[:2] >= (9, 2):
				pass
			else:
				raise
		else:
			if sys.platform in ('win32','win64'):
				self.fail("silent_mode unexpectedly supported on windows")
			elif self.cluster.installation.version_info[:2] >= (9, 2):
				self.fail("silent_mode unexpectedly supported on PostgreSQL >=9.2")

	@unittest.skipIf(default_installation is None, "no installation provided by environment")
	def testSuperPassword(self):
		self.init(
			user = 'test',
			password = 'secret',
			logfile = sys.stdout,
		)
		self.start_cluster()
		c = self.cluster.connection(
			user='test',
			password='secret',
			database='template1',
		)
		with c:
			self.assertEqual(c.prepare('select 1').first(), 1)

	@unittest.skipIf(default_installation is None, "no installation provided by environment")
	def testNoParameters(self):
		"""
		Simple init and drop.
		"""
		self.init()
		self.start_cluster()

if __name__ == '__main__':
	from types import ModuleType
	this = ModuleType("this")
	this.__dict__.update(globals())
	unittest.main(this)



##
# .python.command - Python command emulation module.
##
"""
Create and Execute Python Commands
==================================

The purpose of this module is to simplify the creation of a Python command
interface. Normally, one would want to do this if there is a *common* need
for a certain Python environment that may be, at least, partially initialized
via command line options. A notable case would be a Python environment with a
database connection whose connection parameters came from the command line. That
is, Python + command line driven configuration.

The module also provides an extended interactive console that provides backslash
commands for editing and executing temporary files. Use ``python -m
pythoncommand`` to try it out.

Simple usage::

	import sys
	import os
	import optparse
	import pythoncommand as pycmd

	op = optparse.OptionParser(
		"%prog [options] [script] [script arguments]",
		version = '1.0',
	)
	op.disable_interspersed_args()

	# Basically, the standard -m and -c. (Some additional ones for fun)
	op.add_options(pycmd.default_optparse_options)

	co, ca = op.parse_args(args[1:])

	# This initializes an execution instance which gathers all the information
	# about the code to be ran when ``pyexe`` is called.
	pyexe = pycmd.Execution(ca,
		context = getattr(co, 'python_context', ()),
		loader = getattr(co, 'python_main', None),
	)

	# And run it. Any exceptions will be printed via print_exception.
	rv = pyexe()
	sys.exit(rv)
"""
import os
import sys
import re
import code
import types
import optparse
import subprocess
import contextlib

from gettext import gettext as _
from traceback import print_exception

from pkgutil import get_loader as module_loader

class single_loader(object):
	"""
	used for "loading" string modules(think -c)
	"""
	def __init__(self, source):
		self.source = source

	def get_filename(self, fullpath):
		if fullpath == self.source:
			return '<command>'

	def get_code(self, fullpath):
		if fullpath == self.source:
			return compile(self.source, '<command>', 'exec')

	def get_source(self, fullpath):
		if fullpath == self.source:
			return self.source

class file_loader(object):
	"""
	used for "loading" scripts
	"""
	def __init__(self, filepath, fileobj = None):
		self.filepath = filepath
		if fileobj is not None:
			self._source = fileobj.read()

	def get_filename(self, fullpath):
		if fullpath == self.filepath:
			return self.filepath

	def get_source(self, fullpath):
		if fullpath == self.filepath:
			return self._read()

	def _read(self):
		if hasattr(self, '_source'):
			return self._source
		f = open(self.filepath)
		try:
			return f.read()
		finally:
			f.close()

	def get_code(self, fullpath):
		if fullpath != self.filepath:
			return
		return compile(self._read(), self.filepath, 'exec')

def extract_filepath(x):
	if x.startswith('file://'):
		return x[7:]
	return None

def extract_module(x):
	if x.startswith('module:'):
		return x[7:]
	return None

module_loader_descriptor = (
	'Python module', module_loader, extract_module
)
file_loader_descriptor = (
	'Python script', file_loader, extract_filepath
)
single_loader_descriptor = (
	'Python command', single_loader, lambda x: x
)

_directory = (
	module_loader_descriptor,
	file_loader_descriptor,
)
directory = list(_directory)

def find_loader(ident, dir = directory):
	for x in dir:
		xid = x[2](ident)
		if xid is not None:
			return x

##
# optparse options
##

def append_context(option, opt_str, value, parser):
	"""
	Add some context to the execution of the Python code using
	loader module's directory list of loader descriptions.

	If no loader can be found, assume it's a Python command.
	"""
	pc = getattr(parser.values, option.dest, None) or []
	if not pc:
		setattr(parser.values, option.dest, pc)
	ldesc = find_loader(value)
	if ldesc is None:
		ldesc = single_loader_descriptor
	pc.append((value, ldesc))

def set_python_main(option, opt_str, value, parser):
	"""
	Set the main Python code; after contexts are initialized, main is ran.
	"""
	main = (value, option.python_loader)
	setattr(parser.values, option.dest, main)
	# only terminate parsing if not interspersing arguments
	if not parser.allow_interspersed_args:
		parser.rargs.insert(0, '--')

context = optparse.make_option(
	'-C', '--context',
	help = _('Python context code to run[file://,module:,<code>]'),
	dest = 'python_context',
	action = 'callback',
	callback = append_context,
	type = 'str'
)

module = optparse.make_option(
	'-m',
	help = _('Python module to run as script(__main__)'),
	dest = 'python_main',
	action = 'callback',
	callback = set_python_main,
	type = 'str'
)
module.python_loader = module_loader_descriptor

command = optparse.make_option(
	'-c',
	help = _('Python expression to run(__main__)'),
	dest = 'python_main',
	action = 'callback',
	callback = set_python_main,
	type = 'str'
)
command.python_loader = single_loader_descriptor

default_optparse_options = [
	context, module, command,
]

class ExtendedConsole(code.InteractiveConsole):
	"""
	Console subclass providing some convenient backslash commands.
	"""
	def __init__(self, *args, **kw):
		import tempfile
		self.mktemp = tempfile.mktemp
		import shlex
		self.split = shlex.split
		code.InteractiveConsole.__init__(self, *args, **kw)

		self.bsc_map = {}
		self.temp_files = {}
		self.past_buffers = []

		self.register_backslash(r'\?', self.showhelp, "Show this help message.")
		self.register_backslash(r'\set', self.bs_set,
			"Configure environment variables. \set without arguments to show all")
		self.register_backslash(r'\E', self.bs_E,
			"Edit a file or a temporary script.")
		self.register_backslash(r'\i', self.bs_i,
			"Execute a Python script within the interpreter's context.")
		self.register_backslash(r'\e', self.bs_e,
			"Edit and Execute the file directly in the context.")
		self.register_backslash(r'\x', self.bs_x,
			"Execute the Python command within this process.")

	def interact(self, *args, **kw):
		self.showhelp(None, None)
		return super().interact(*args,**kw)

	def showtraceback(self):
		e, v, tb = sys.exc_info()
		sys.last_type, sys.last_value, sys.last_traceback = e, v, tb
		print_exception(e, v, tb.tb_next or tb)

	def register_backslash(self, bscmd, meth, doc):
		self.bsc_map[bscmd] = (meth, doc)

	def execslash(self, line):
		"""
		If push() gets a line that starts with a backslash, execute
		the command that the backslash sequence corresponds to.
		"""
		cmd = line.split(None, 1)
		cmd.append('')
		bsc = self.bsc_map.get(cmd[0])
		if bsc is None:
			self.write("ERROR: unknown backslash command: %s%s"%(cmd, os.linesep))
		else:
			return bsc[0](cmd[0], cmd[1])

	def showhelp(self, cmd, arg):
		i = list(self.bsc_map.items())
		i.sort(key = lambda x: x[0])
		helplines = os.linesep.join([
			'  %s%s%s' %(
				x[0], ' ' * (8 - len(x[0])), x[1][1]
			) for x in i
		])
		self.write("Backslash Commands:%s%s%s" %(
			os.linesep*2, helplines, os.linesep*2
		))

	def bs_set(self, cmd, arg):
		"""
		Set a value in the interpreter's environment.
		"""
		if arg:
			for x in self.split(arg):
				if '=' in x:
					k, v = x.split('=', 1)
					os.environ[k] = v
					self.write("%s=%s%s" %(k, v, os.linesep))
				elif x:
					self.write("%s=%s%s" %(x, os.environ.get(x, ''), os.linesep))
		else:
			for k,v in os.environ.items():
				self.write("%s=%s%s" %(k, v, os.linesep))

	def resolve_path(self, path, dont_create = False):
		"""
		Get the path of the given string; if the path is not
		absolute and does not contain path separators, identify
		it as a temporary file.
		"""
		if not os.path.isabs(path) and not os.path.sep in path:
			# clean it up to avoid typos
			path = path.strip().lower()
			tmppath = self.temp_files.get(path)
			if tmppath is None:
				if dont_create is False:
					tmppath = self.mktemp(
						suffix = '.py',
						prefix = '_console_%s_' %(path,)
					)
					self.temp_files[path] = tmppath
				else:
					return path
			return tmppath
		return path

	def execfile(self, filepath):
		src = open(filepath)
		try:
			try:
				co = compile(src.read(), filepath, 'exec')
			except SyntaxError:
				co = None
				print_exception(*sys.exc_info())
		finally:
			src.close()
		if co is not None:
			try:
				exec(co, self.locals, self.locals)
			except:
				e, v, tb = sys.exc_info()
				print_exception(e, v, tb.tb_next or tb)

	def editfiles(self, filepaths):
		sp = list(filepaths)
		# ;)
		sp.insert(0, os.environ.get('EDITOR', 'vi'))
		return subprocess.call(sp)

	def bs_i(self, cmd, arg):
		'execute the files'
		for x in self.split(arg) or ('',):
			p = self.resolve_path(x, dont_create = True)
			self.execfile(p)

	def bs_E(self, cmd, arg):
		'edit the files, but *only* edit them'
		self.editfiles([self.resolve_path(x) for x in self.split(arg) or ('',)])

	def bs_e(self, cmd, arg):
		'edit *and* execute the files'
		filepaths = [self.resolve_path(x) for x in self.split(arg) or ('',)]
		self.editfiles(filepaths)
		for x in filepaths:
			self.execfile(x)

	def bs_x(self, cmd, arg):
		rv = -1
		if len(cmd) > 1:
			a = self.split(arg)
			a.insert(0, '\\x')
			try:
				rv = command(argv = a)
			except SystemExit as se:
				rv = se.code
			self.write("[Return Value: %d]%s" %(rv, os.linesep))

	def push(self, line):
		# Has to be a ps1 context.
		if not self.buffer and line.startswith('\\'):
			try:
				self.execslash(line)
			except:
				# print the exception, but don't raise.
				e, v, tb = sys.exc_info()
				print_exception(e, v, tb.tb_next or tb)
		else:
			return code.InteractiveConsole.push(self, line)

@contextlib.contextmanager
def postmortem(funcpath):
	if not funcpath:
		yield None
	else:
		pm = funcpath.split('.')
		attr = pm.pop(-1)
		modpath = '.'.join(pm)
		try:
			m = __import__(modpath, fromlist = modpath)
			pmobject = getattr(m, attr, None)
		except ValueError:
			pmobject = None

			sys.stderr.write(
				"%sERROR: no object at %r for postmortem%s"%(
					os.linesep, funcpath, os.linesep
				)
			)
		try:
			yield None
		except:
			try:
				sys.last_type, sys.last_value, sys.last_traceback = sys.exc_info()
				pmobject()
			except:
				sys.stderr.write(
					"[Exception raised by Postmortem]" + os.linesep
				)
				print_exception(*sys.exc_info())
			raise

class Execution(object):
	"""
	Given argv and context make an execution instance that, when called, will
	execute the configured Python code.

	This class provides the ability to identify what the main part of the
	execution of the configured Python code. For instance, shall it execute a
	console, the file that the first argument points to, a -m option module
	appended to the python_context option value, or the code given within -c?
	"""
	def __init__(self,
		args, context = (),
		main = None,
		loader = None,
		stdin = sys.stdin
	):
		"""
		args
			The arguments passed to the script; usually sys.argv after being
			processed by optparse(ca).
		context
			A list of loader descriptors that will be used to establish the
			context of __main__ module.
		main
			Overload to explicitly state what main is. None will cause the
			class to attempt to fill in the attribute using 'args' and other
			system objects like sys.stdin.
		"""
		self.args = args
		self.context = context and list(context) or ()

		if main is not None:
			self.main = main
		elif loader is not None:
			# Main explicitly stated, resolve the path and the loader
			path, ldesc = loader
			ltitle, rloader, xpath = ldesc
			l = rloader(path)
			if l is None:
				raise ImportError(
					"%s %r does not exist or cannot be read" %(
						ltitle, path
					)
				)
			self.main = (path, l)
		# If there are args, but no main, run the first arg.
		elif args:
			fp = self.args[0]
			f = open(fp)
			try:
				l = file_loader(fp, fileobj = f)
			finally:
				f.close()
			self.main = (self.args[0], l)
			self.args = self.args[1:]
		# There is no main, no loader, and no args.
		# If stdin is not a tty, use stdin as the main file.
		elif not stdin.isatty():
			l = file_loader('<stdin>', fileobj = stdin)
			self.main = ('<stdin>', l)
		# tty and no "main".
		else:
			# console
			self.main = (None, None)
		self.reset_module__main__()

	def reset_module__main__(self):
		mod = types.ModuleType('__main__')
		mod.__builtins__ = __builtins__
		mod.__package__ = None
		self.module__main__ = mod
		path = getattr(self.main[1], 'fullname', None)
		if path is not None:
			mod.__package__ = '.'.join(path.split('.')[:-1])

	def _call(self,
		console = ExtendedConsole,
		context = None
	):
		"""
		Initialize the context and run main in the given locals
		(Note: tramples on sys.argv, __main__ in sys.modules)
		(Use __call__ instead)
		"""
		sys.modules['__main__'] = self.module__main__
		md = self.module__main__.__dict__

		# Establish execution context in the locals;
		# iterate over all the loaders in self.context and
		for path, ldesc in self.context:
			ltitle, loader, xpath = ldesc
			rpath = xpath(path)
			li = loader(rpath)
			if li is None:
				sys.stderr.write(
					"%s %r does not exist or cannot be read%s" %(
						ltitle, rpath, os.linesep
					)
				)
				return 1
			try:
				code = li.get_code(rpath)
			except:
				print_exception(*sys.exc_info())
				return 1
			self.module__main__.__file__ = getattr(
				li, 'get_filename', lambda x: x
			)(rpath)
			self.module__main__.__loader__ = li
			try:
				exec(code, md, md)
			except:
				e, v, tb = sys.exc_info()
				print_exception(e, v, tb.tb_next or tb)
				return 1

		if self.main == (None, None):
			# It's interactive.
			sys.argv = self.args or ['<console>']

			# Use readline if available
			try:
				import readline
			except ImportError:
				pass

			ic = console(locals = md)
			try:
				ic.interact()
			except SystemExit as e:
				return e.code
			return 0
		else:
			# It's ultimately a code object.
			path, loader = self.main
			self.module__main__.__file__ = getattr(
				loader, 'get_filename', lambda x: x
			)(path)
			sys.argv = list(self.args)
			sys.argv.insert(0, self.module__main__.__file__)
			try:
				code = loader.get_code(path)
			except:
				print_exception(*sys.exc_info())
				return 1

			rv = 0
			exe_exception = False
			try:
				if context is not None:
					with context:
						try:
							exec(code, md, md)
						except:
							exe_exception = True
							raise
				else:
					try:
						exec(code, md, md)
					except:
						exe_exception = True
						raise

			except SystemExit as e:
				# Assume it's an exe_exception as anything ran in `context`
				# shouldn't cause an exception.
				rv = e.code
				e, v, tb = sys.exc_info()
				sys.last_type = e
				sys.last_value = v
				sys.last_traceback = (tb.tb_next or tb)
			except:
				if exe_exception is False:
					raise
				rv = 1
				e, v, tb = sys.exc_info()
				print_exception(e, v, tb.tb_next or tb)
				sys.last_type = e
				sys.last_value = v
				sys.last_traceback = (tb.tb_next or tb)

			return rv

	def __call__(self, *args, **kw):
		storage = (
			sys.modules.get('__context__'),
			sys.modules.get('__main__'),
			sys.argv,
			os.environ.copy(),
		)
		try:
			return self._call(*args, **kw)
		finally:
			sys.modules['__context__'], \
			sys.modules['__main__'], \
			sys.argv, os.environ = storage

	def get_main_source(self):
		"""
		Get the execution's "__main__" source. Useful for configuring
		environmental options derived from "magic" lines.
		"""
		path, loader = self.main
		if path is not None:
			return loader.get_source(path)

def command_execution(argv = sys.argv):
	'create an execution using the given argv'
	# The pwd should be in the path for python commands.
	# setuptools' console_scripts appear to strip this out.
	if '' not in sys.path:
		sys.path.insert(0, '')

	op = optparse.OptionParser(
		"%prog [options] [script] [script arguments]",
		version = '1.0',
	)
	op.disable_interspersed_args()
	op.add_options(default_optparse_options)
	co, ca = op.parse_args(argv[1:])

	return Execution(ca,
		context = getattr(co, 'python_context', ()),
		loader = getattr(co, 'python_main', None),
	)

def command(argv = sys.argv):
	return command_execution(argv = argv)(
		context = postmortem(os.environ.get('PYTHON_POSTMORTEM'))
	)

if __name__ == '__main__':
	sys.exit(command())



##
# .test.test_bytea_codec
##
import unittest
import struct
from ..encodings import bytea

byte = struct.Struct('B')

class test_bytea_codec(unittest.TestCase):
	def testDecoding(self):
		for x in range(255):
			c = byte.pack(x)
			b = c.decode('bytea')
			# normalize into octal escapes
			if c == b'\\' and b == "\\\\":
				b = "\\" + oct(b'\\'[0])[2:]
			elif not b.startswith("\\"):
				b = "\\" + oct(ord(b))[2:]
			if int(b[1:], 8) != x:
				self.fail(
					"bytea encoding failed at %d; encoded %r to %r" %(x, c, b,)
				)

	def testEncoding(self):
		self.assertEqual('bytea'.encode('bytea'), b'bytea')
		self.assertEqual('\\\\'.encode('bytea'), b'\\')
		self.assertRaises(ValueError, '\\'.encode, 'bytea')
		self.assertRaises(ValueError, 'foo\\'.encode, 'bytea')
		self.assertRaises(ValueError, r'foo\0'.encode, 'bytea')
		self.assertRaises(ValueError, r'foo\00'.encode, 'bytea')
		self.assertRaises(ValueError, r'\f'.encode, 'bytea')
		self.assertRaises(ValueError, r'\800'.encode, 'bytea')
		self.assertRaises(ValueError, r'\7f0'.encode, 'bytea')
		for x in range(255):
			seq = ('\\' + oct(x)[2:].lstrip('0').rjust(3, '0'))
			dx = ord(seq.encode('bytea'))
			if dx != x:
				self.fail(
					"generated sequence failed to map back; current is %d, " \
					"rendered %r, transformed to %d" %(x, seq, dx)
				)

if __name__ == '__main__':
	unittest.main()



##
# types.io.stdlib_xml_etree
##
try:
	import xml.etree.cElementTree as etree
except ImportError:
	import xml.etree.ElementTree as etree
from .. import XMLOID
from ...python.functools import Composition as compose

oid_to_type = {
	XMLOID: etree.ElementTree,
}

def xml_unpack(xmldata, XML = etree.XML):
	try:
		return XML(xmldata)
	except Exception:
		# try it again, but return the sequence of children.
		return tuple(XML('<x>' + xmldata + '</x>'))

if not hasattr(etree, 'tostringlist'):
	# Python 3.1 support.
	def xml_pack(xml, tostr = etree.tostring, et = etree.ElementTree,
		str = str, isinstance = isinstance, tuple = tuple
	):
		if isinstance(xml, str):
			# If it's a string, encode and return.
			return xml
		elif isinstance(xml, tuple):
			# If it's a tuple, encode and return the joined items.
			# We do not accept lists here--emphasizing lists being used for ARRAY
			# bounds.
			return ''.join((x if isinstance(x, str) else tostr(x) for x in xml))
		return tostr(xml)

	def xml_io_factory(typoid, typio, c = compose):
		return (
			c((xml_pack, typio.encode)),
			c((typio.decode, xml_unpack)),
			etree.ElementTree,
		)
else:
	# New etree tostring API.
	def xml_pack(xml, encoding, encoder,
		tostr = etree.tostring, et = etree.ElementTree,
		str = str, isinstance = isinstance, tuple = tuple,
	):
		if isinstance(xml, bytes):
			return xml
		if isinstance(xml, str):
			# If it's a string, encode and return.
			return encoder(xml)
		elif isinstance(xml, tuple):
			# If it's a tuple, encode and return the joined items.
			# We do not accept lists here--emphasizing lists being used for ARRAY
			# bounds.
			##
			# 3.2
			# XXX: tostring doesn't include declaration with utf-8?
			x = b''.join(
				x.encode('utf-8') if isinstance(x, str) else
				tostr(x, encoding = "utf-8")
				for x in xml
			)
		else:
			##
			# 3.2
			# XXX: tostring doesn't include declaration with utf-8?
			x = tostr(xml, encoding = "utf-8")
		if encoding in ('utf8','utf-8'):
			return x
		else:
			return encoder(x.decode('utf-8'))

	def xml_io_factory(typoid, typio, c = compose):
		def local_xml_pack(x, encoder = typio.encode, typio = typio, xml_pack = xml_pack):
			return xml_pack(x, typio.encoding, encoder)
		return (local_xml_pack, c((typio.decode, xml_unpack)), etree.ElementTree,)

oid_to_io = {
	XMLOID : xml_io_factory
}



##
# .iri
##
"""
Parse and serialize PQ IRIs.

PQ IRIs take the form::

	pq://user:pass@host:port/database?setting=value&setting2=value2

IPv6 is supported via the standard representation::

	pq://[::1]:5432/database

Driver Parameters:

	pq://user@host/?[driver_param]=value&[other_param]=value?server_setting=val
"""
from .resolved import riparse as ri
from .string import split_ident

from operator import itemgetter
get0 = itemgetter(0)
del itemgetter

import re
escape_path_re = re.compile('[%s]' %(re.escape(ri.unescaped + ','),))

def structure(l, fieldproc = ri.unescape):
	"""
	Create a clientparams dictionary from a parsed RI.
	"""
	res = []
	for d in l:
		scheme = d.get('scheme', 'pq').lower()
		if scheme not in {'pq', 'postgres', 'postgresql', 'og', 'opengauss'}:
			raise ValueError("PQ-IRI scheme is not 'pq', 'postgres', 'postgresql', 'og' or 'opengauss'")
		if scheme in {'og', 'opengauss'}:
			# recover opengauss scheme to pq
			d['scheme'] = 'pq'

		cpd = {
			k : fieldproc(v) for k, v in d.items()
			if k not in ('path', 'fragment', 'query', 'host', 'scheme')
		}

		path = d.get('path')
		frag = d.get('fragment')
		query = d.get('query')
		host = d.get('host')

		if host is not None:
			if host.startswith('[') and host.endswith(']'):
				host = host[1:-1]
				if host.startswith('unix:'):
					cpd['unix'] = host[len('unix:'):].replace(':','/')
				else:
					cpd['host'] = host
			else:
				cpd['host'] = fieldproc(host)

		if path:
			# Only state the database field's existence if the first path is non-empty.
			if path[0]:
				cpd['database'] = path[0]
			path = path[1:]
			if path:
				cpd['path'] = path

		settings = {}
		if query:
			if hasattr(query, 'items'):
				qiter = query.items()
			else:
				qiter = query
			for k, v in qiter:
				if k.startswith('[') and k.endswith(']'):
					k = k[1:-1]
					if k != 'settings' and k not in cpd:
						cpd[fieldproc(k)] = fieldproc(v)
				elif k:
					settings[fieldproc(k)] = fieldproc(v)
				# else: ignore empty query keys

		if frag:
			settings['search_path'] = [
				fieldproc(x) for x in frag.split(',')
			]

		if settings:
			cpd['settings'] = settings
		res.append(cpd)

	return res

def construct_path(x, re = escape_path_re):
	"""
	Join a path sequence using ',' and escaping ',' in the pieces.
	"""
	return ','.join((re.sub(ri.re_pct_encode, y) for y in x))

def construct(x, obscure_password = False):
	"""
	Construct a RI dictionary from a clientparams dictionary.
	"""
	# the rather exhaustive settings choreography is due to
	# a desire to allow the search_path to be appended in the fragment
	settings = x.get('settings')
	no_path_settings = None
	search_path = None
	if settings:
		if isinstance(settings, dict):
			siter = settings.items()
			search_path = settings.get('search_path')
		else:
			siter = list(settings)
			search_path = [(k,v) for k,v in siter if k == 'search_path']
			search_path.append((None,None))
			search_path = search_path[-1][1]
		no_path_settings = [(k,v) for k,v in siter if k != 'search_path']
		if not no_path_settings:
			no_path_settings = None

	# It could be a string search_path, split if it is.
	if search_path is not None and isinstance(search_path, str):
		search_path = split_ident(search_path, sep = ',')

	port = None
	if 'unix' in x:
		host = '[unix:' + x['unix'].replace('/',':') + ']'
		# ignore port.. it's a mis-config.
	elif 'host' in x:
		host = x['host']
		if ':' in host:
			host = '[' + host + ']'
		port = x.get('port')
	else:
		host = None
		port = x.get('port')

	path = []
	if 'database' in x:
		path.append(x['database'])
	if 'path' in x:
		path.extend(x['path'] or ())

	password = x.get('password')
	if obscure_password and password is not None:
		password = '***'
	driver_params = list({
		'[' + k + ']' : str(v) for k,v in x.items()
		if k not in (
			'user', 'password', 'port', 'database', 'ssl',
			'path', 'host', 'unix', 'ipv','settings'
		)
	}.items())
	driver_params.sort(key=get0)

	return (
		'pqs' if x.get('ssl', False) is True else 'pq',
		# netloc: user:pass@host[:port]
		ri.unsplit_netloc((
			x.get('user'),
			password,
			host,
			None if 'port' not in x else str(x['port'])
		)),
		None if not path else '/'.join([
			ri.escape_path_re.sub(path_comp, '/')
			for path_comp in path
		]),
		(ri.construct_query(driver_params) if driver_params else None)
		if no_path_settings is None else (
			ri.construct_query(
				driver_params + no_path_settings
			)
		),
		None if search_path is None else construct_path(search_path),
	)

def parse(s, fieldproc = ri.unescape):
	"""
	Parse a Postgres IRI into a dictionary object list.
	"""
	return structure(
		# In ri.parse, don't unescape the parsed values as our sub-structure
		# uses the escape mechanism in IRIs to specify literal separator
		# characters.
		ri.parse(s, fieldproc = str),
		fieldproc = fieldproc
	)

def serialize(x, obscure_password = False):
	"""
	Return a Postgres IRI from a dictionary object.
	"""
	return ri.unsplit(construct(x, obscure_password = obscure_password))

if __name__ == '__main__':
	import sys
	for x in sys.argv[1:]:
		print("{src} -> {parsed!r} -> {serial}".format(
			src = x,
			parsed = parse(x),
			serial = serialize(parse(x))
		))



##
# .python.socket - additional tools for working with sockets
##
import sys
import os
import random
import socket
import errno
import ssl

__all__ = ['find_available_port', 'SocketFactory']

class SocketFactory(object):
	"""
	Object used to create a socket and connect it.

	This is, more or less, a specialized partial() for socket creation.

	Additionally, it provides methods and attributes for abstracting
	exception management on socket operation.
	"""

	timeout_exception = socket.timeout
	fatal_exception = socket.error
	try_again_exception = socket.error

	def timed_out(self, err) -> bool:
		return err.__class__ is self.timeout_exception

	@staticmethod
	def try_again(err, codes = (errno.EAGAIN, errno.EINTR, errno.EWOULDBLOCK, errno.ETIMEDOUT)) -> bool:
		"""
		Does the error indicate that the operation should be tried again?

		More importantly, the connection is *not* dead.
		"""
		errno = getattr(err, 'errno', None)
		if errno is None:
			return False
		return errno in codes

	@classmethod
	def fatal_exception_message(typ, err) -> (str, None):
		"""
		If the exception was fatal to the connection,
		what message should be given to the user?
		"""
		if typ.try_again(err):
			return None
		return getattr(err, 'strerror', '<strerror not present>')

	def secure(self, socket : socket.socket) -> ssl.SSLSocket:
		"""
		Secure a socket with SSL.
		"""
		if self.socket_secure is not None:
			return ssl.wrap_socket(socket, **self.socket_secure)
		else:
			return ssl.wrap_socket(socket)

	def __call__(self, timeout = None):
		s = socket.socket(*self.socket_create)
		try:
			s.settimeout(float(timeout) if timeout is not None else None)
			s.connect(self.socket_connect)
			s.settimeout(None)
		except Exception:
			s.close()
			raise
		return s

	def __init__(self,
		socket_create,
		socket_connect,
		socket_secure = None,
	):
		self.socket_create = socket_create
		self.socket_connect = socket_connect
		self.socket_secure = socket_secure

	def __str__(self):
		return 'socket' + repr(self.socket_connect)

def find_available_port(
	interface = 'localhost',
	address_family = socket.AF_INET,
):
	"""
	Find an available port on the given interface for the given address family.
	"""

	port = None
	s = socket.socket(address_family, socket.SOCK_STREAM,)
	try:
		s.bind(('localhost', 0))
		port = s.getsockname()[1]
	finally:
		s.close()

	return port



##
# python.functools
##
import sys
from .decorlib import method

def rsetattr(attr, val, ob):
	"""
	setattr() and return `ob`. Different order used to allow easier partial
	usage.
	"""
	setattr(ob, attr, val)
	return ob

try:
	from ..port.optimized import rsetattr
except ImportError:
	pass

class Composition(tuple):
	def __call__(self, r):
		for x in self:
			r = x(r)
		return r

	try:
		from ..port.optimized import compose
		__call__ = method(compose)
		del compose
	except ImportError:
		pass

try:
	# C implementation of the tuple processors.
	from ..port.optimized import process_tuple, process_chunk
except ImportError:
	def process_tuple(procs, tup, exception_handler, len = len, tuple = tuple, cause = None):
		"""
		Call each item in `procs` with the corresponding
		item in `tup` returning the result as `type`.

		If an item in `tup` is `None`, don't process it.

		If a give transformation failes, call the given exception_handler.
		"""
		i = len(procs)
		if len(tup) != i:
			raise TypeError(
				"inconsistent items, %d processors and %d items in row" %(
					i, len(tup)
				)
			)
		r = [None] * i
		try:
			for i in range(i):
				ob = tup[i]
				if ob is None:
					continue
				r[i] = procs[i](ob)
		except Exception:
			cause = sys.exc_info()[1]

		if cause is not None:
			exception_handler(cause, procs, tup, i)
			raise RuntimeError("process_tuple exception handler failed to raise")
		return tuple(r)

	def process_chunk(procs, tupc, fail, process_tuple = process_tuple):
		return [process_tuple(procs, x, fail) for x in tupc]



##
# .encodings.bytea
##
'PostgreSQL bytea encoding and decoding functions'
import codecs
import struct
import sys

ord_to_seq = {
	i : \
		"\\" + oct(i)[2:].rjust(3, '0') \
		if not (32 < i < 126) else r'\\' \
		if i == 92 else chr(i)
	for i in range(256)
}

if sys.version_info[:2] >= (3, 3):
	# Subscripting memory in 3.3 returns byte as an integer, not as a bytestring
	def decode(data):
		return ''.join(map(ord_to_seq.__getitem__, (data[x] for x in range(len(data)))))
else:
	def decode(data):
		return ''.join(map(ord_to_seq.__getitem__, (data[x][0] for x in range(len(data)))))

def encode(data):
	diter = ((data[i] for i in range(len(data))))
	output = []
	next = diter.__next__
	for x in diter:
		if x == "\\":
			try:
				y = next()
			except StopIteration:
				raise ValueError("incomplete backslash sequence")
			if y == "\\":
				# It's a backslash, so let x(\) be appended.
				x = ord(x)
			elif y.isdigit():
				try:
					os = ''.join((y, next(), next()))
				except StopIteration:
					# requires three digits
					raise ValueError("incomplete backslash sequence")
				try:
					x = int(os, base = 8)
				except ValueError:
					raise ValueError("invalid bytea octal sequence '%s'" %(os,))
			else:
				raise ValueError("invalid backslash follow '%s'" %(y,))
		else:
			x = ord(x)
		output.append(x)
	return struct.pack(str(len(output)) + 'B', *output)

class Codec(codecs.Codec):
	'bytea codec'
	def encode(data, errors = 'strict'):
		return (encode(data), len(data))
	encode = staticmethod(encode)

	def decode(data, errors = 'strict'):
		return (decode(data), len(data))
	decode = staticmethod(decode)

class StreamWriter(Codec, codecs.StreamWriter): pass
class StreamReader(Codec, codecs.StreamReader): pass

bytea_codec = (Codec.encode, Codec.decode, StreamReader, StreamWriter)
codecs.register(lambda x: x == 'bytea' and bytea_codec or None)



##
# test.perf_copy_io - Copy I/O: To and From performance
##
import os, sys, random, time

if __name__ == '__main__':
	with open('/usr/share/dict/words', mode='brU') as wordfile:
		Words = wordfile.readlines()
else:
	Words = [b'/usr/share/dict/words', b'is', b'read', b'in', b'__main__']
wordcount = len(Words)
random.seed()

def getWord():
	"extract a random word from ``Words``"
	return Words[random.randrange(0, wordcount)].strip()

def testSpeed(tuples = 50000 * 3):
	sqlexec("CREATE TEMP TABLE _copy "
	"(i int, t text, mt text, ts text, ty text, tx text);")
	try:
		Q = prepare("COPY _copy FROM STDIN")
		size = 0
		def incsize(data):
			'count of bytes'
			nonlocal size
			size += len(data)
			return data
		sys.stderr.write("preparing data(%d tuples)...\n" %(tuples,))

		# Use an LC to avoid the Python overhead involved with a GE
		data = [incsize(b'\t'.join((
			str(x).encode('ascii'), getWord(), getWord(),
			getWord(), getWord(), getWord()
		)))+b'\n' for x in range(tuples)]

		sys.stderr.write("starting copy...\n")
		start = time.time()
		copied_in = Q.load_rows(data)
		duration = time.time() - start
		sys.stderr.write(
			"COPY FROM STDIN Summary,\n " \
			"copied tuples: %d\n " \
			"copied bytes: %d\n " \
			"duration: %f\n " \
			"average tuple size(bytes): %f\n " \
			"average KB per second: %f\n " \
			"average tuples per second: %f\n" %(
				tuples, size, duration,
				size / tuples,
				size / 1024 / duration,
				tuples / duration, 
			)
		)
		Q = prepare("COPY _copy TO STDOUT")
		start = time.time()
		c = 0
		for rows in Q.chunks():
			c += len(rows)
		duration = time.time() - start
		sys.stderr.write(
			"COPY TO STDOUT Summary,\n " \
			"copied tuples: %d\n " \
			"duration: %f\n " \
			"average KB per second: %f\n " \
			"average tuples per second: %f\n " %(
				c, duration,
				size / 1024 / duration,
				tuples / duration, 
			)
		)
	finally:
		sqlexec("DROP TABLE _copy")

if __name__ == '__main__':
	testSpeed()



##
# .string
##
"""
String split and join operations for dealing with literals and identifiers.

Notably, the functions in this module are intended to be used for simple
use-cases. It attempts to stay away from "real" parsing and simply provides
functions for common needs, like the ability to identify unquoted portions of a
query string so that logic or transformations can be applied to only unquoted
portions. Scanning for statement terminators, or safely interpolating
identifiers.

All functions deal with strict quoting rules.
"""
import re

def escape_literal(text):
	"""
	Replace every instance of ' with ''.
	"""
	return text.replace("'", "''")

def quote_literal(text):
	"""
	Escape the literal and wrap it in [single] quotations.
	"""
	return "'" + text.replace("'", "''") + "'"

def escape_ident(text):
	"""
	Replace every instance of " with "".
	"""
	return text.replace('"', '""')

def needs_quoting(text):
	return not (text and not text[0].isdecimal() and text.replace('_', 'a').isalnum())

def quote_ident(text):
	"""
	Replace every instance of '"' with '""' *and* place '"' on each end.
	"""
	return '"' + text.replace('"', '""') + '"'

def quote_ident_if_needed(text):
	"""
	If needed, replace every instance of '"' with '""' *and* place '"' on each end.
	Otherwise, just return the text.
	"""
	return quote_ident(text) if needs_quoting(text) else text

quote_re = re.compile(r"""(?xu)
	E'(?:''|\\.|[^'])*(?:'|$)          (?# Backslash escapes E'str')
|	'(?:''|[^'])*(?:'|$)               (?# Regular literals 'str')
|	"(?:""|[^"])*(?:"|$)               (?# Identifiers "str")
|	(\$(?:[^0-9$]\w*)?\$).*?(?:\1|$)   (?# Dollar quotes $$str$$)
""")

def split(text):
	"""
	split the string up by into non-quoted and quoted portions. Zero and even
	numbered indexes are unquoted portions, while odd indexes are quoted
	portions.

	Unquoted portions are regular strings, whereas quoted portions are
	pair-tuples specifying the quotation mechanism and the content thereof.

	>>> list(split("select $$foobar$$"))
	['select ', ('$$', 'foobar'), '']

	If the split ends on a quoted section, it means the string's quote was not
	terminated. Subsequently, there will be an even number of objects in the
	list.

	Quotation errors are detected, but never raised. Rather it's up to the user
	to identify the best course of action for the given split.
	"""
	lastend = 0
	re = quote_re
	scan = re.scanner(text)
	match = scan.search()
	while match is not None:
		# text preceding the quotation
		yield text[lastend:match.start()]
		# the dollar quote, if any
		dq = match.groups()[0]
		if dq is not None:
			endoff = len(dq)
			quote = dq
			end = quote
		else:
			endoff = 1
			q = text[match.start()]
			if q == 'E':
				quote = "E'"
				end = "'"
			else:
				end = quote = q

		# If the end is not the expected quote, it consumed
		# the end. Be sure to check that the match's end - end offset
		# is *not* the start, ie an empty quotation at the end of the string.
		if text[match.end()-endoff:match.end()] != end \
		or match.end() - endoff == match.start():
			yield (quote, text[match.start()+len(quote):])
			break
		else:
			yield (quote, text[match.start()+len(quote):match.end()-endoff])

		lastend = match.end()
		match = scan.search()
	else:
		# balanced quotes, yield the rest
		yield text[lastend:]

def unsplit(splitted_iter):
	"""
	catenate a split string. This is needed to handle the special
	cases created by pg.string.split(). (Run-away quotations, primarily)
	"""
	s = ''
	quoted = False
	i = iter(splitted_iter)
	endq = ''
	for x in i:
		s += endq + x
		try:
			q, qtext = next(i)
			s += q + qtext
			if q == "E'":
				endq = "'"
			else:
				endq = q
		except StopIteration:
			break
	return s

def split_using(text, quote, sep = '.', maxsplit = -1):
	"""
	split the string on the seperator ignoring the separator in quoted areas.

	This is only useful for simple quoted strings. Dollar quotes, and backslash
	escapes are not supported.
	"""
	escape = quote * 2
	esclen = len(escape)
	offset = 0
	tl = len(text)
	end = tl
	# Fast path: No quotes? Do a simple split.
	if quote not in text:
		return text.split(sep, maxsplit)
	l = []

	while len(l) != maxsplit:
		# Look for the separator first
		nextsep = text.find(sep, offset)
		if nextsep == -1:
			# it's over. there are no more seps
			break
		else:
			# There's a sep ahead, but is there a quoted section before it?
			nextquote = text.find(quote, offset, nextsep)
			while nextquote != -1:
				# Yep, there's a quote before the sep;
				# need to eat the escaped portion.
				nextquote = text.find(quote, nextquote + 1,)
				while nextquote != -1:
					if text.find(escape, nextquote, nextquote+esclen) != nextquote:
						# Not an escape, so it's the end.
						break
					# Look for another quote past the escape quote.
					nextquote = text.find(quote, nextquote + 2)
				else:
					# the sep was located in the escape, and
					# the escape consumed the rest of the string.
					nextsep = -1
					break

				nextsep = text.find(sep, nextquote + 1)
				if nextsep == -1:
					# it's over. there are no more seps
					# [likely they were consumed by the escape]
					break
				nextquote = text.find(quote, nextquote + 1, nextsep)
			if nextsep == -1:
				break

			l.append(text[offset:nextsep])
			offset = nextsep + 1
	l.append(text[offset:])
	return l

def split_ident(text, sep = ',', quote = '"', maxsplit = -1):
	"""
	Split a series of identifiers using the specified separator.
	"""
	nr = []
	for x in split_using(text, quote, sep = sep, maxsplit = maxsplit):
		x = x.strip()
		if x.startswith('"'):
			if not x.endswith('"'):
				raise ValueError(
					"unterminated identifier quotation", x
				)
			else:
				nr.append(x[1:-1].replace('""', '"'))
		elif needs_quoting(x):
			raise ValueError(
				"non-ident characters in unquoted identifier", x
			)
		else:
			# postgres implies a lower, so to stay consistent
			# with it on qname joins, lower the unquoted identifier now.
			nr.append(x.lower())
	return nr

def split_qname(text, maxsplit = -1):
	"""
	Call to .split_ident() with a '.' sep parameter.
	"""
	return split_ident(text, maxsplit = maxsplit, sep = '.')

def qname(*args):
	"""
	Quote the identifiers and join them using '.'.
	"""
	return '.'.join([quote_ident(x) for x in args])

def qname_if_needed(*args):
	return '.'.join([quote_ident_if_needed(x) for x in args])

def split_sql(sql, sep = ';'):
	"""
	Given SQL, safely split using the given separator.
	Notably, this yields fully split text. This should be used instead of
	split_sql_str() when quoted sections need be still be isolated.

	>>> list(split_sql('select $$1$$ AS "foo;"; select 2;'))
	[['select ', ('$$', '1'), ' AS ', ('"', 'foo;'), ''], (' select 2',), ['']]
	"""
	i = iter(split(sql))
	cur = []
	for part in i:
		sections = part.split(sep)

		if len(sections) < 2:
			cur.append(part)
		else:
			cur.append(sections[0])
			yield cur
			for x in sections[1:-1]:
				yield (x,)
			cur = [sections[-1]]
		try:
			cur.append(next(i))
		except StopIteration:
			break
	if cur:
		yield cur

def split_sql_str(sql, sep = ';'):
	"""
	Identical to split_sql but yields unsplit text.

	>>> list(split_sql_str('select $$1$$ AS "foo;"; select 2;'))
	['select $$1$$ AS "foo;"', ' select 2', '']
	"""
	for x in split_sql(sql, sep = sep):
		yield unsplit(x)



##
# test.test_optimized
##
import unittest
import struct
import sys
from ..port import optimized
from ..python.itertools import interlace

def pack_tuple(*data,
	packH = struct.Struct("!H").pack,
	packL = struct.Struct("!L").pack
):
	return packH(len(data)) + b''.join((
		packL(len(x)) + x if x is not None else b'\xff\xff\xff\xff'
		for x in data
	))

tuplemessages = (
	(b'D', pack_tuple(b'foo', b'bar')),
	(b'D', pack_tuple(b'foo', None, b'bar')),
	(b'N', b'fee'),
	(b'D', pack_tuple(b'foo', None, b'bar')),
	(b'D', pack_tuple(b'foo', b'bar')),
)

class test_optimized(unittest.TestCase):
	def test_consume_tuple_messages(self):
		ctm = optimized.consume_tuple_messages
		# expecting a tuple of pairs.
		self.assertRaises(TypeError, ctm, [])
		self.assertEqual(ctm(()), [])
		# Make sure that the slicing is working.
		self.assertEqual(ctm(tuplemessages), [
			(b'foo', b'bar'),
			(b'foo', None, b'bar'),
		])
		# Not really checking consume here, but we are validating that
		# it's properly propagating exceptions.
		self.assertRaises(ValueError, ctm, ((b'D', b'\xff\xff\xff\xfefoo'),))
		self.assertRaises(ValueError, ctm, ((b'D', b'\x00\x00\x00\x04foo'),))

	def test_parse_tuple_message(self):
		ptm = optimized.parse_tuple_message
		self.assertRaises(TypeError, ptm, "stringzor")
		self.assertRaises(TypeError, ptm, 123)
		self.assertRaises(ValueError, ptm, b'')
		self.assertRaises(ValueError, ptm, b'0')

		notenoughdata = struct.pack('!H', 2)
		self.assertRaises(ValueError, ptm, notenoughdata)

		wraparound = struct.pack('!HL', 2, 10) + (b'0' * 10) + struct.pack('!L', 0xFFFFFFFE)
		self.assertRaises(ValueError, ptm, wraparound)

		oneatt_notenough = struct.pack('!HL', 2, 10) + (b'0' * 10) + struct.pack('!L', 15)
		self.assertRaises(ValueError, ptm, oneatt_notenough)

		toomuchdata = struct.pack('!HL', 1, 3) + (b'0' * 10)
		self.assertRaises(ValueError, ptm, toomuchdata)

		class faketup(tuple):
			def __new__(subtype, geeze):
				r = tuple.__new__(subtype, ())
				r.foo = geeze
				return r
		zerodata = struct.pack('!H', 0)
		r = ptm(zerodata)
		self.assertRaises(AttributeError, getattr, r, 'foo')
		self.assertRaises(AttributeError, setattr, r, 'foo', 'bar')
		self.assertEqual(len(r), 0)

	def test_process_tuple(self):
		def funpass(procs, tup, col):
			pass
		pt = optimized.process_tuple
		# tuple() requirements
		self.assertRaises(TypeError, pt, "foo", "bar", funpass)
		self.assertRaises(TypeError, pt, (), "bar", funpass)
		self.assertRaises(TypeError, pt, "foo", (), funpass)
		self.assertRaises(TypeError, pt, (), ("foo",), funpass)

	def test_pack_tuple_data(self):
		pit = optimized.pack_tuple_data
		self.assertEqual(pit((None,)), b'\xff\xff\xff\xff')
		self.assertEqual(pit((None,)*2), b'\xff\xff\xff\xff'*2)
		self.assertEqual(pit((None,)*3), b'\xff\xff\xff\xff'*3)
		self.assertEqual(pit((None,b'foo')), b'\xff\xff\xff\xff\x00\x00\x00\x03foo')
		self.assertEqual(pit((None,b'')), b'\xff\xff\xff\xff\x00\x00\x00\x00')
		self.assertEqual(pit((None,b'',b'bar')), b'\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x03bar')
		self.assertRaises(TypeError, pit, 1)
		self.assertRaises(TypeError, pit, (1,))
		self.assertRaises(TypeError, pit, ("",))

	def test_int2(self):
		d = b'\x00\x01'
		rd = b'\x01\x00'
		s = optimized.swap_int2_unpack(d)
		n = optimized.int2_unpack(d)
		sd = optimized.swap_int2_pack(1)
		nd = optimized.int2_pack(1)
		if sys.byteorder == 'little':
			self.assertEqual(1, s)
			self.assertEqual(256, n)
			self.assertEqual(d, sd)
			self.assertEqual(rd, nd)
		else:
			self.assertEqual(1, n)
			self.assertEqual(256, s)
			self.assertEqual(d, nd)
			self.assertEqual(rd, sd)
		self.assertRaises(OverflowError, optimized.swap_int2_pack, 2**15)
		self.assertRaises(OverflowError, optimized.int2_pack, 2**15)
		self.assertRaises(OverflowError, optimized.swap_int2_pack, (-2**15)-1)
		self.assertRaises(OverflowError, optimized.int2_pack, (-2**15)-1)

	def test_int4(self):
		d = b'\x00\x00\x00\x01'
		rd = b'\x01\x00\x00\x00'
		s = optimized.swap_int4_unpack(d)
		n = optimized.int4_unpack(d)
		sd = optimized.swap_int4_pack(1)
		nd = optimized.int4_pack(1)
		if sys.byteorder == 'little':
			self.assertEqual(1, s)
			self.assertEqual(16777216, n)
			self.assertEqual(d, sd)
			self.assertEqual(rd, nd)
		else:
			self.assertEqual(1, n)
			self.assertEqual(16777216, s)
			self.assertEqual(d, nd)
			self.assertEqual(rd, sd)
		self.assertRaises(OverflowError, optimized.swap_int4_pack, 2**31)
		self.assertRaises(OverflowError, optimized.int4_pack, 2**31)
		self.assertRaises(OverflowError, optimized.swap_int4_pack, (-2**31)-1)
		self.assertRaises(OverflowError, optimized.int4_pack, (-2**31)-1)

	def test_int8(self):
		d = b'\x00\x00\x00\x00\x00\x00\x00\x01'
		rd = b'\x01\x00\x00\x00\x00\x00\x00\x00'
		s = optimized.swap_int8_unpack(d)
		n = optimized.int8_unpack(d)
		sd = optimized.swap_int8_pack(1)
		nd = optimized.int8_pack(1)
		if sys.byteorder == 'little':
			self.assertEqual(0x1, s)
			self.assertEqual(0x100000000000000, n)
			self.assertEqual(d, sd)
			self.assertEqual(rd, nd)
		else:
			self.assertEqual(0x1, n)
			self.assertEqual(0x100000000000000, s)
			self.assertEqual(d, nd)
			self.assertEqual(rd, sd)
		self.assertEqual(optimized.swap_int8_pack(-1), b'\xFF\xFF\xFF\xFF'*2)
		self.assertEqual(optimized.int8_pack(-1), b'\xFF\xFF\xFF\xFF'*2)
		self.assertRaises(OverflowError, optimized.swap_int8_pack, 2**63)
		self.assertRaises(OverflowError, optimized.int8_pack, 2**63)
		self.assertRaises(OverflowError, optimized.swap_int8_pack, (-2**63)-1)
		self.assertRaises(OverflowError, optimized.int8_pack, (-2**63)-1)
		# edge I/O
		int8_max = ((2**63) - 1)
		int8_min = (-(2**63))
		swap_max = optimized.swap_int8_pack(int8_max)
		max = optimized.int8_pack(int8_max)
		swap_min = optimized.swap_int8_pack(int8_min)
		min = optimized.int8_pack(int8_min)
		self.assertEqual(optimized.swap_int8_unpack(swap_max), int8_max)
		self.assertEqual(optimized.int8_unpack(max), int8_max)
		self.assertEqual(optimized.swap_int8_unpack(swap_min), int8_min)
		self.assertEqual(optimized.int8_unpack(min), int8_min)

	def test_uint2(self):
		d = b'\x00\x01'
		rd = b'\x01\x00'
		s = optimized.swap_uint2_unpack(d)
		n = optimized.uint2_unpack(d)
		sd = optimized.swap_uint2_pack(1)
		nd = optimized.uint2_pack(1)
		if sys.byteorder == 'little':
			self.assertEqual(1, s)
			self.assertEqual(256, n)
			self.assertEqual(d, sd)
			self.assertEqual(rd, nd)
		else:
			self.assertEqual(1, n)
			self.assertEqual(256, s)
			self.assertEqual(d, nd)
			self.assertEqual(rd, sd)
		self.assertRaises(OverflowError, optimized.swap_uint2_pack, -1)
		self.assertRaises(OverflowError, optimized.uint2_pack, -1)
		self.assertRaises(OverflowError, optimized.swap_uint2_pack, 2**16)
		self.assertRaises(OverflowError, optimized.uint2_pack, 2**16)
		self.assertEqual(optimized.uint2_pack(2**16-1), b'\xFF\xFF')
		self.assertEqual(optimized.swap_uint2_pack(2**16-1), b'\xFF\xFF')

	def test_uint4(self):
		d = b'\x00\x00\x00\x01'
		rd = b'\x01\x00\x00\x00'
		s = optimized.swap_uint4_unpack(d)
		n = optimized.uint4_unpack(d)
		sd = optimized.swap_uint4_pack(1)
		nd = optimized.uint4_pack(1)
		if sys.byteorder == 'little':
			self.assertEqual(1, s)
			self.assertEqual(16777216, n)
			self.assertEqual(d, sd)
			self.assertEqual(rd, nd)
		else:
			self.assertEqual(1, n)
			self.assertEqual(16777216, s)
			self.assertEqual(d, nd)
			self.assertEqual(rd, sd)
		self.assertRaises(OverflowError, optimized.swap_uint4_pack, -1)
		self.assertRaises(OverflowError, optimized.uint4_pack, -1)
		self.assertRaises(OverflowError, optimized.swap_uint4_pack, 2**32)
		self.assertRaises(OverflowError, optimized.uint4_pack, 2**32)
		self.assertEqual(optimized.uint4_pack(2**32-1), b'\xFF\xFF\xFF\xFF')
		self.assertEqual(optimized.swap_uint4_pack(2**32-1), b'\xFF\xFF\xFF\xFF')

	def test_uint8(self):
		d = b'\x00\x00\x00\x00\x00\x00\x00\x01'
		rd = b'\x01\x00\x00\x00\x00\x00\x00\x00'
		s = optimized.swap_uint8_unpack(d)
		n = optimized.uint8_unpack(d)
		sd = optimized.swap_uint8_pack(1)
		nd = optimized.uint8_pack(1)
		if sys.byteorder == 'little':
			self.assertEqual(0x1, s)
			self.assertEqual(0x100000000000000, n)
			self.assertEqual(d, sd)
			self.assertEqual(rd, nd)
		else:
			self.assertEqual(0x1, n)
			self.assertEqual(0x100000000000000, s)
			self.assertEqual(d, nd)
			self.assertEqual(rd, sd)
		self.assertRaises(OverflowError, optimized.swap_uint8_pack, -1)
		self.assertRaises(OverflowError, optimized.uint8_pack, -1)
		self.assertRaises(OverflowError, optimized.swap_uint8_pack, 2**64)
		self.assertRaises(OverflowError, optimized.uint8_pack, 2**64)
		self.assertEqual(optimized.uint8_pack((2**64)-1), b'\xFF\xFF\xFF\xFF'*2)
		self.assertEqual(optimized.swap_uint8_pack((2**64)-1), b'\xFF\xFF\xFF\xFF'*2)

if __name__ == '__main__':
	from types import ModuleType
	this = ModuleType("this")
	this.__dict__.update(globals())
	unittest.main(this)
